-module(wpc_pov).

-export([init/0,menu/2,dialog/2,command/2,init_pref/0,
  get_var/1,set_var/2,erase_var/1,get_pref/2,set_pref/1]).

-import(lists, [foldl/3,map/2,foreach/2,reverse/1,seq/2,flat_length/1,append/2]).

-import(kayos_utils,[find_executable/1,cr_to_nl/1,db/1,db/2,get_the_right_slashes/2,char_to_int/1]).

-include("e3d.hrl").
-include("e3d_image.hrl").
-include("pov.hrl").
-include_lib("kernel/include/file.hrl").

init() ->
  init_pref (),
  set_var (rendering, false),
  true.

menu(Item, Menu) ->
  pov_ui:menu(Item, Menu).
dialog(Type,Res) ->
  pov_ui:dialog(Type,Res).

command({file,{export,{?TAG,A}}}, St) ->
  command_file(export, A, St);
command({file,{export_selected,{?TAG,A}}}, St) ->
  command_file(export_selected, A, St);
command({file,{render,{?TAG,A}}}, St) ->
  command_file(render, A, St);
command({file,{?TAG_RENDER,Data}}, St) ->
  command_file(?TAG_RENDER, Data, St);
command({edit,{plugin_preferences,?TAG}}, St) ->
  pov_ui:pref_edit(St);
command(_Spec, _St) ->
  next.

init_pref() ->
  {OsGroup,_Os}=os:type(),
  set_var (pov_osgroup, OsGroup),

  RendererPath = case filename:pathtype("povray") of
    absolute -> "povray";
    _ ->
      case find_executable("povray") of
        false ->
	  PovPath = case get_var(pov_path) of
	    undefined -> undefined;
	    PPath -> [34] ++ PPath ++ [34] end, % 34 == quotation mark "
	  case find_executable(PovPath) of false -> false; undefined -> false; Path -> Path end;
        Path -> Path end end,

  case OsGroup of
    unix ->
      set_var (pov_tmpdir, "/tmp"),
      set_var (pov_version, povray),
      set_var (pov_path, "/usr/local/bin/povray"),
      set_var (pov_render_options, ""),
      case get_pref(dialogs, ?DEF_DIALOGS) of
        auto -> set_var(dialogs, RendererPath);
        enabled -> set_var(dialogs, RendererPath);
        disabled -> set_var(dialogs, false)
      end,
      case get_pref(dialogs, ?DEF_DIALOGS) of
        auto -> set_var(renderer, RendererPath);
        enabled -> set_var(renderer, RendererPath);
        disabled -> set_var(renderer, false)
      end;
    win32 ->
      set_var(renderer,false),
      set_var(dialogs, true);
    _ ->
      io:put_chars ("unknown OS, please report\n contact details at http://www.midcoast.com.au/~rgcoy\n") end,

  ok.

command_file(render, Attr0, St) when is_list (Attr0) ->
  set_pref(Attr0),
  case get_var(rendering) of
    false ->
      Attr = pov_ui:set_additional_preferences(Attr0),
      UseSSS = get_pref(use_sss, ?DEF_USE_SSS),
      SubDiv = case UseSSS of
        false -> get_pref(subdivisions, ?DEF_SUBDIVISIONS);
        true -> 0 end,
      wpa:export(props(?TAG, SubDiv), fun_export_2(attr(St,
        [{?TAG_RENDER,true},{pov_tmpdir,get_var(pov_tmpdir)},{use_sss,UseSSS}|Attr]
        )), St);
    _RenderFile ->
      wpa:error("Already rendering.") end;
command_file(render, Ask, _St) when is_atom(Ask) ->
  wpa:dialog(Ask, "Povray Render Options", pov_ui:export_dialog(render),
                   fun(Attr) -> {file,{render,{?TAG,Attr}}} end);
command_file(?TAG_RENDER, {Pid,Result,Ack}, _St) ->
  Rendering = set_var(rendering, false),
  Pid ! Ack,
  case Rendering of
    false -> keep;
    RenderFile ->
      case Result of
        load_image ->
	  io:format("Loading rendered image~n"),
	  load_image(RenderFile);
	  ok -> keep;
	  _ -> wpa:error("Rendering error")
      end
  end;
command_file(Op,Attr0,St) when is_list(Attr0) ->
  set_pref(Attr0),
  Attr = pov_ui:set_additional_preferences(Attr0),
  UseSSS = get_pref(use_sss, ?DEF_USE_SSS),
  SubDiv = case UseSSS of
    false -> get_pref(subdivisions, ?DEF_SUBDIVISIONS);
    true -> 0 end,
  wpa:Op(props(?TAG, SubDiv), fun_export_2(attr(St, [{use_sss,UseSSS}|Attr])), St);
command_file(Op, Ask, _St) when is_atom(Ask) ->
  wpa:dialog(Ask, "Povray Export Options", pov_ui:export_dialog(export),
                   fun(Attr) -> {file,{Op,{?TAG,Attr}}} end).

props(tga, SubDiv) ->
  [{ext,".tga"},{ext_desc,"Targa File"},{subdivisions,SubDiv}];
props(?TAG, SubDiv) ->
  [{ext,".pov"},{ext_desc,"Povray File"},{subdivisions,SubDiv}].

attr(St, Attr) ->
  [Aim,Dist,Az,El,Track,Fov] =
    wpa:camera_info([aim,distance_to_aim,azimuth,elevation,tracking,fov]),
  CameraInfo = #camera_info{aim=Aim,distance_to_aim=Dist,azimuth=Az,
    elevation=El,tracking=Track,fov=Fov},
  [CameraInfo,{lights,wpa:lights(St)}|Attr].
%attr(St, Attr) ->
%  [{lights, wpa:lights(St)}|Attr].

load_image(Filename) ->
  case wpa:image_read([{filename,Filename}, {alignment,1}]) of
    #e3d_image{}=Image ->
      Id = wings_image:new("Rendered (Povray)", Image),
      wings_image:window(Id),
      keep;
    _ ->
      wpa:error("No Image Rendered")
    end.

fun_export_2(Attr) ->
  fun (Filename, Contents) ->
    case catch export(Attr, Filename, Contents) of
      ok -> ok;
      Error -> io:format("ERROR: Failed to export:~n~p~n", [Error]), {error, "Failed to export"} end end.

export(Attr0, FileName, Contents) ->
  RenderImg = proplists:get_value(?TAG_RENDER, Attr0, false),
  Tmpdir0 = proplists:get_value(pov_tmpdir, Attr0, none),
  BaseFile = filename:rootname (filename:basename (FileName)),
  {Tmpdir,RenderFile0} = case RenderImg of
    true ->
      {A,B,C} = erlang:now(),
      Tmpdir1 = lists:flatten(io_lib:format("~s/wingspov-~s-~w-~w-~w/",[Tmpdir0,os:getpid(),A,B,C])),
      {Tmpdir1, Tmpdir1 ++ BaseFile ++ ".tga"};
    false ->
      {none, BaseFile ++ ".tga"} end,
  RenderFile = get_the_right_slashes(RenderFile0,get_var(pov_osgroup)),
  Attr = [{pov_tmpdir,Tmpdir}|Attr0],
  pov_exp:export(FileName,Contents,Attr),
  Renderer = get_var(renderer),
  case {Renderer,RenderImg} of
    {false,_} -> ok;
    {_,false} -> ok;
    _ ->
      Options0 = get_var(pov_render_options),
      Options = " " ++ Options0 ++
                " Width=" ++ integer_to_list(proplists:get_value(frame_width,Attr,320)) ++
                " Height=" ++ integer_to_list(proplists:get_value(frame_height,Attr,240)) ++
                " +FC Output_File_Name=" ++ [34] ++ RenderFile ++ [34] ++ " ", % 34 = "\""
      LoadImage = proplists:get_value(load_image, Attr, ?DEF_LOAD_IMAGE),
      Parent = self(),
      spawn_link(
      fun () ->
        file:delete(RenderFile),
	render(Renderer, Options, FileName, Tmpdir, Parent, LoadImage)
      end),
      db("RenderFile", RenderFile),
      set_var(rendering, RenderFile),
    ok
  end.

render(Renderer, Options, FileName, Tmpdir, Parent, LoadImage) ->
  process_flag(trap_exit, true),
  DirName = filename:dirname(FileName),
  BaseName = filename:basename(FileName),
  InputFile = get_the_right_slashes(FileName,get_var(pov_osgroup)),
  Cmd = [34] ++ Renderer ++ [34] ++ " " ++ Options ++ " Input_File_Name=" ++ [34] ++ InputFile ++ [34], % 34 = "\""
  PortOpts = [{line,8},{cd,DirName},eof,exit_status,stderr_to_stdout],
  io:format("Rendering Job started ~p:~n>~s~n", [self(),Cmd]),
  case catch open_port({spawn,Cmd}, PortOpts) of
    Port when port(Port) ->
      Result = render_job(Port),
      render_done(FileName, Parent, Tmpdir, Result, LoadImage);
    {'EXIT',Reason} ->
      render_done(FileName, Parent, Tmpdir, {error, Reason}, LoadImage)
  end.

render_done(Filename, Parent, Tmpdir, ExitStatus, LoadImage) ->
  Status = case ExitStatus of
    0 -> ok;
    undefined -> ok;
    _ -> ExitStatus
  end,
  Result = case {Status,LoadImage} of
    {ok,true} ->
      load_image;
    {ok,false} ->
      ok;
    {{error,_},_} ->
      Status;
    {Error,_} ->
      {error,Error}
  end,
  io:format("~nRendering Job returned: ~p~n",[ExitStatus]),
  db("OOO Tmpdir", Tmpdir),
  case Tmpdir of
    none -> file:delete(Filename);
    _ ->
      case file:list_dir(Tmpdir) of
      {ok, Dirlist} ->
        db("TMP_DIR_ooo", Dirlist),
        foreach(fun(Tmpfile) -> db("Tmpfile",Tmpfile),file:delete(Tmpfile) end, Dirlist),
        file:del_dir(Tmpdir);
      _ ->
        io:put_chars("failed to delete temporary files\n")
      end
    end,
  render_cleanup(Parent, make_ref(), Result).

render_cleanup(Parent, Ref, Result) ->
  Ack = {Parent,ack,Ref},
  Command = {file,{?TAG_RENDER,{self(),Result,Ack}}},
  Parent ! {timeout,Ref,{event,{action,Command}}},
  receive Ack ->
    io:format("Rendering Job ready~n"),
    ok
  after 5000 ->
    render_cleanup(Parent, Ref, Result)
  end.

render_job(Port) ->
  receive
    {Port,eof} ->
      receive
        {Port,{exit_status,ExitStatus}} ->
          ExitStatus
      after 1 ->
          undefined
      end;
    {port,{exit_status,ExitStatus}} ->
      receive
        {Port, eof} ->
          ok after 1 -> ok end,
      ExitStatus;
    {Port, {data,{Tag,Data}}} ->
      io:put_chars(cr_to_nl(Data)),
      case Tag of eol -> io:nl(); _ -> ok end,
      render_job(Port);
    {'EXIT',Port,Reason} ->
      {error,Reason};
    Other ->
      io:format("WARNING: Unexpected at ~s:~p:~p~n", [?MODULE_STRING,?LINE,Other]),
      render_job(Port)
    end.

get_var(Name) ->
    get({?POV_MODULE,Name}).

set_var(Name, undefined) ->
    erase_var(Name);
set_var(Name, Value) ->
    put({?POV_MODULE,Name}, Value).

erase_var(Name) ->
    erase({?POV_MODULE,Name}).

set_pref(Attr) ->
    wpa:pref_set(?POV_MODULE, Attr).

get_pref(Key, Def) ->
    wpa:pref_get(?POV_MODULE, Key, Def).
